#!/usr/bin/env python
# coding = utf-8
"""功能简要说明
作者：dengqingyong
邮箱：yu12377@163.com
时间：2021/6/15 16:57
"""
import json
import hmac
import base64
import hashlib
import logging
import requests
from deng import time as time_utils
from deng import my_http as http_utils


logger = logging.getLogger("DengUtils")

url_model = "https://open.feishu.cn/open-apis/bot/v2/hook/{0}"
headers = {"Content-Type": "application/json; charset=utf-8"}


def get_sign(secret: str) -> tuple:
    timestamp = time_utils.get_timestamp(length=10)
    raw_str = f"{timestamp}\n{secret}".encode("utf-8")
    hmac_code = hmac.new(raw_str, "".encode("utf-8"), digestmod=hashlib.sha256).digest()
    sign = base64.b64encode(hmac_code).decode("utf-8")
    return sign, timestamp


def send_text_alert_message(message: str, receiver: dict, app_name: str = ""):
    """发送文本消息
    :param app_name: 消息来源应用
    :param message: 消息内容
    :param receiver: dict, 必须包含id与secret字段，消息接收者——飞书群自定义消息机器人
    """
    try:
        sign, timestamp = get_sign(receiver["secret"])
        url = url_model.format(receiver["id"])
        if app_name:
            message += f"。消息来源：{app_name}"
        payload = {
            "sign": sign,
            "timestamp": timestamp,
            "msg_type": "text",
            "content": {"text": message},
        }

        res = requests.post(url, headers=headers, json=payload)
        if (
            res.status_code == 200
            and "application/json" in res.headers.get("Content-Type")
            and res.json()["StatusCode"] == 0
        ):
            http_utils.format_output(payload)
            logger.info("消息发送成功")
        else:
            http_utils.format_output(res, level="error")
            logger.error("消息发送失败")
    except Exception as _error:
        logger.exception(_error)


def send_post_message(title: str, content_list: list, receiver: dict, app_name: str = ""):
    """发送富文本消息"""
    try:
        sign, timestamp = get_sign(receiver["secret"])
        url = url_model.format(receiver["id"])
        if app_name:
            title += f"({app_name})"
        payload = {
            "sign": sign,
            "timestamp": timestamp,
            "msg_type": "post",
            "content": {
                "post": {
                    "zh_cn": {
                        "title": title,
                        "content": content_list,
                    }
                }
            },
        }

        try:
            res = requests.post(url, headers=headers, json=payload)
        except Exception as _e:
            logger.exception(_e)
            logger.error(f"访问FeiShu接口失败")
            logger.info(payload)
            raise
        else:
            try:
                if (
                    res.status_code == 200
                    and "application/json" in res.headers.get("Content-Type")
                    and res.json()["StatusCode"] == 0
                ):
                    http_utils.format_output(payload)
                    logger.info(f"消息发送成功")
                else:
                    http_utils.format_output(res, level="error")
                    logger.error(f"消息发送失败")
            except Exception as _e:
                logger.exception(_e)
                logger.error(res.text)
                logger.info(payload)
    except Exception as _error:
        logger.exception(_error)


if __name__ == "__main__":
    pass



import os
import json
import requests
from urllib.parse import urlparse, unquote
from requests.cookies import RequestsCookieJar
from requests.structures import CaseInsensitiveDict

from . import logger
from .utils import get_caller_info, MyJSONEncoder


def parser_cookies_by_headers(headers: CaseInsensitiveDict):
    cookies = RequestsCookieJar()
    cookie_str = headers.get("Cookie", "")
    if cookie_str:
        for cookie in cookie_str.strip().split(";"):
            key, value = cookie.strip().split("=")
            cookies.set(key.strip(), value.strip())
    return cookies


def check_response(
    response: requests.Response,
    operate="",
    ignore_error=True,
    echo=False,
    caller_info: dict = None,
    disable_echo: bool = False,
) -> bool:
    """检查response是否成功
    :param response: Requests的请求返回对象
    :param operate: str, 请求操作描述，为空时会自动根据函数调用栈定位到源函数，获取其描述信息
    :param ignore_error: bool, 为True时即使请求不成功也不抛错，直接返回Response原始对象
    :param echo: bool, 强制输出日志，优化级低于disable_echo
    :param caller_info: dict, 调用源信息，优先级低于operate
    :param disable_echo: bool, 为True时禁用显示请求报文，优先级最高
    """
    if response.status_code == 200:
        content_type = response.headers.get("content-type", "")
        if content_type and "json" in content_type:
            if not operate:
                _info = caller_info or get_caller_info(2)
                operate = f"调用{_info['output']})"
            if response.json().get("code") in (0, 200, "200"):
                status = True
            elif response.json().get("success") is True:
                status = True
            else:
                msg = f"{operate}失败，错误详情：{response.text}"
                if ignore_error:
                    status = False
                    logger.error(msg)
                else:
                    if disable_echo is False:
                        format_output(response, level="error", depth=3)
                    raise ValueError(msg)

            # 根据请求结果与调试标志输出
            if not status or echo or os.environ.get("ALWAYS_ECHO_RESULT"):
                if status:
                    level = "debug"
                else:
                    level = "error"
                if disable_echo is False:
                    format_output(response, level=level, depth=3)
            return status
        else:
            return True
    else:
        if ignore_error:
            if disable_echo is False:
                format_output(response, level="warning", depth=3)
            return False
        else:
            if disable_echo is False:
                format_output(response, level="error", depth=3)
            raise ValueError(response.text)


def format_output(obj, level="debug", depth=2, caller_info: dict = None):
    def __output_response(__obj: requests.Response):
        _logger(f"请求地址：{__obj.request.method}, {__obj.request.url}")
        _logger(
            f"请求头部：{json.dumps(dict(__obj.request.headers), ensure_ascii=False, indent=4, cls=MyJSONEncoder)}"
        )
        if __obj.request.body:
            if "application/json" in __obj.request.headers.get("Content-Type", ""):
                if isinstance(__obj.request.body, bytes):
                    body = json.loads(__obj.request.body.decode("utf-8"))
                else:
                    body = json.loads(__obj.request.body)
                _logger(f"请求内容：{json.dumps(body, ensure_ascii=False, indent=4, cls=MyJSONEncoder)}")
            elif "multipart/form-data" in __obj.request.headers.get("Content-Type", ""):
                if isinstance(__obj.request.body, bytes):
                    _logger(f"请求内容：可能包含文件内容，忽略不显示")
                else:
                    _logger(f"请求内容：{str(__obj.request.body)}")
            else:
                if isinstance(__obj.request.body, bytes):
                    _logger(f"请求内容：{__obj.request.body.decode('utf-8')}")
                else:
                    _logger(f"请求内容：{__obj.request.body}")
        _logger(f"响应首行：{__obj.status_code}, {__obj.reason}, {__obj.url}, 耗时：{__obj.elapsed}")
        _logger(f"响应头部：{json.dumps(dict(__obj.headers), ensure_ascii=False, indent=4, cls=MyJSONEncoder)}")
        if __obj.text.strip():
            if "application/json" in __obj.headers.get("Content-Type", ""):
                _logger(f"响应内容：{json.dumps(__obj.json(), ensure_ascii=False, indent=4, cls=MyJSONEncoder)}")
            else:
                if "attachment" in __obj.headers.get("Content-Disposition", ""):
                    _logger(f"响应内容：包含文件内容，忽略不显示")
                else:
                    _logger(f"响应内容：{__obj.text}")

    _info = caller_info or get_caller_info(depth)
    _logger = getattr(logger, level)
    _logger(f"格式化输出调用源：{_info['output']}")
    try:
        if isinstance(obj, (dict, list, tuple)):
            _logger(json.dumps(obj, ensure_ascii=False, indent=4, cls=MyJSONEncoder))
        elif isinstance(obj, requests.Response):
            if obj.history:
                for __obj in obj.history:
                    __output_response(__obj)
            __output_response(obj)
        else:
            _logger(obj)
    except Exception as _e:
        logger.exception(_e)
        logger.warning("格式法输出对象出错，忽略")
        _logger(f"type={type(obj)}, value={str(obj)}")


def clean_null_key_for_dict(my_dict: dict):
    """将字典中第一层的空值清除"""
    if isinstance(my_dict, dict):
        for key, value in list(my_dict.items()):
            if value is None:
                my_dict.pop(key)
    return my_dict


def parse_url_to_dict(url: str, _all=True, unquote_count: int = 0) -> dict:
    """将url解析成字典"""
    payload = dict()
    url = unquote_url(url, unquote_count)
    url_obj = urlparse(url)
    if _all:
        payload["scheme"] = url_obj.scheme
        payload["netloc"] = url_obj.netloc
        payload["path"] = url_obj.path
        payload["params"] = url_obj.params
        payload["query"] = url_obj.query
        payload["fragment"] = url_obj.fragment

    query_dict = query_str_to_dict(url_obj.query)
    fragment_dict = query_str_to_dict(url_obj.fragment)
    payload.update(query_dict)
    payload.update(fragment_dict)
    return payload


def pop_key_from_dict(my_dict: dict, key, default=None):
    if key in my_dict:
        value = my_dict.pop(key)
    else:
        value = default
    return value


def query_str_to_dict(_str: str) -> dict:
    """将URL查询字符串转换成dict"""
    payload = dict()
    if _str:
        _str = unquote(_str)

        while "?" in _str:
            _key, _str = _str.split("?", 1)
            logger.debug(f"剥离{_key}")

        for item in _str.split("&"):
            try:
                key, values = item.split("=")
            except ValueError as e:
                logger.warning("拆包时出现异常，可能是字符串中存在多个【=】导致，启用备用方案拆包：忽略第2个及以后的【=】字符")
                logger.warning("原始字符串：{}".format(item))
                key, values = item.split("=", 1)
            payload[key] = values
    return payload


def unquote_url(url, unquote_count: int = 0) -> str:
    """解析URL"""
    if unquote_count > 0:
        # 按指定的循环次数解码
        for _ in range(unquote_count):
            url = unquote(url)
    else:
        # 自动解码
        while True:
            _before_url = url
            url = unquote(url)
            if _before_url == url:
                break
    return url



import os
import re
import json
import time
from pathlib import Path
from typing import Union, List

from . import logger, file as file_utils


def get_engine_version_info(deputy_abs_path: Union[str, Path]) -> dict:
    """获取引擎版本信息"""
    file_utils.check_path_is_exits(deputy_abs_path, path_type="file")
    version_path = Path(deputy_abs_path).parent / "version.txt"
    file_utils.check_path_is_exits(version_path, path_type="file")
    version_info = dict()
    with open(version_path, encoding="utf-8") as _file:
        try:
            _file_version_info = json.load(_file)
        except Exception as e:
            raise ValueError(f"版本文件解析出错：{version_path}")

    version_info["arch"] = _file_version_info["InstructionSet"]
    # 兼容三段版本号与四段版本号
    version_info["raw_version"] = _file_version_info["Version"]
    _temp_version_list = version_info["raw_version"].split(".")
    if len(_temp_version_list) == 4:
        version_info["version"] = ".".join(_temp_version_list[:-1])
    elif len(_temp_version_list) == 3:
        version_info["version"] = _file_version_info["Version"]
    else:
        raise ValueError(f"引擎版本出现未知情况：{_file_version_info}")
    version_info["language"] = _file_version_info.get("Language", "zh-cn")
    version_info["build_date"] = _file_version_info["Build"]

    product_name_list = _file_version_info["Product"].lower().split()
    version_info["edition"] = product_name_list[-1]
    if len(product_name_list) == 3:
        version_info["package"] = product_name_list[1]
    else:
        version_info["package"] = "creator"

    return version_info


def get_flow_info(flow_path: Union[Path, str]):
    """从流程flow文件中读取流程信息"""
    if isinstance(flow_path, str):
        flow_path = Path(flow_path)
    if flow_path.is_dir():
        flow_file_abs_path = flow_path / "main.prj"
        if not flow_file_abs_path.exists():
            flow_file_abs_path = flow_path / f"{flow_path.name}.flow"
    else:
        flow_file_abs_path = flow_path
    file_utils.check_path_is_exits(flow_file_abs_path, path_type="file")
    with open(flow_file_abs_path, encoding="utf-8") as _file:
        flow_config = json.load(_file)
    return flow_config


def update_flow_info(flow_path: str, flow_info: dict):
    """更新流程flow文件信息"""
    file_utils.check_path_is_exits(flow_path)
    flow_path = Path(flow_path)
    if flow_path.is_dir():
        flow_file_path = flow_path / f"{flow_path.name}.flow"
    else:
        flow_file_path = flow_path
    file_utils.check_path_is_exits(flow_file_path, path_type="file")
    # 备份flow文件
    file_utils.copy_to_target(flow_file_path, str(flow_file_path) + str(time.time()) + ".bak")
    with open(flow_file_path, mode="w", encoding="utf-8") as _file:
        json.dump(flow_info, _file, ensure_ascii=False, indent=4)


def compare_version_number(version_a: str, version_b: str, level: int = 3) -> int:
    """获取最后的版本信息
    :param version_a: str, 比较参数1
    :param version_b: str, 比较参数2
    :param level: int, 比较等级，0时全版本比较，1时仅比较第一个大版本，2时比较前2个版本号，3号比较前3个版本号
    """
    a = version_a.split(".")
    b = version_b.split(".")
    if len(a) != len(b):
        raise ValueError(f"版本格式不一致，无法比较：{version_a}，{version_b}")
    for i in range(len(a[:level])):
        res = int(a[i]) - int(b[i])
        if res != 0:
            return res
    return 0


def flow_global_param_convert(flow_params: list, target_type) -> list:
    """流程全局参数转换"""
    new_flow_params = []
    _msg_old = "将flow转换成5.2.0以前版本格式"
    _msg_new = "将flow转换成5.2.0及以后版本格式"
    for param in flow_params:
        if isinstance(param, target_type):
            new_flow_params.append(param)
            continue

        if target_type == str:
            if _msg_old:
                logger.info(_msg_old)
                _msg_old = None
            if param["type"] == "none":
                new_flow_params.append(param["var"])
            else:
                raise ValueError(f"流程中用到了子流程，无法转换成5.2.0以前的版本，不兼容！")
        elif target_type == dict:
            if _msg_new:
                logger.info(_msg_new)
                _msg_new = None
            temp = {"var": param, "type": "none"}
            new_flow_params.append(temp)
    return new_flow_params


def get_next_version(version: str) -> str:
    """获取下一版本号"""
    v = version.split(".")
    last_number = str(int(v.pop()) + 1)
    v.append(last_number)
    return ".".join(v)


def version_increment(version: str, step: int = 10) -> str:
    """获取下一版本号
    版本号如：1.0.0，返回1.0.1
    :param version: 版本号
    :param step: 版本号递增步长
    :return 下一版本号
    """
    # 将版本号字符串拆分为整数列表
    version_parts = list(map(int, version.split('.')))
    length = len(version_parts)

    # 从最后一位开始递增
    for i in range(length - 1, 0, -1):
        version_parts[i] += 1

        # 如果当前位不需要进位（即小于 10），结束循环
        if version_parts[i] < step:
            break
        # 如果需要进位（即达到 10），将当前位设为 0，继续向前进位
        version_parts[i] = 0

    # 如果是最后一位进位导致循环结束，最低位也递增
    else:
        version_parts[0] += 1

    # 将版本号列表转换回字符串形式
    return '.'.join(map(str, version_parts))


def is_ignore(ignore_path: Path, flow_file_path: Path) -> bool:
    """判断是否忽略流程"""
    try:
        file_utils.check_path_is_exits(ignore_path, path_type="file")
    except FileNotFoundError as _e:
        logger.warning(f"忽略列表文件不存在，不忽略任何流程")
        return False

    with open(ignore_path, encoding="utf-8") as _file:
        for _line in _file:
            _line = _line.strip()
            # 跳过空行
            if not _line:
                continue
            # 跳过注释行
            if _line.startswith("//"):
                continue
            # 解析配置行
            if "=" in _line:
                try:
                    _type, _value = _line.split("=", 1)
                    # 去掉=两边的空格
                    _type = _type.strip()
                    _value = _value.strip()
                except Exception as _e:
                    # 跳过无效配置行
                    continue
                else:
                    # 按路径忽略
                    if _type.lower() == "path":
                        _item_ignore_path = ignore_path.parent / _value
                        if _item_ignore_path.exists():
                            # 具体到流程文件
                            if _item_ignore_path == flow_file_path:
                                return True
                            # 按目录忽略
                            if _item_ignore_path in flow_file_path.parents:
                                return True
                    # 按流程名称忽略
                    elif _type.lower() == "name":
                        if flow_file_path.stem == _value:
                            return True
                    else:
                        # 跳过无效配置行
                        continue
            else:
                # 跳过无效配置行
                continue
    # 前面逻辑都未匹配时
    return False


def is_author(author, flow_file_path: Path) -> bool:
    """判断author是否为流程作者"""
    return get_flow_author(flow_file_path) == author


def get_file_list_by_suffix(
    target: Path,
    suffix: str,
    reverse: bool = True,
    depth=2,
    flow_version: str = None,
    flow_name: str = None,
    ignore_path: Path = None,
    only_author: str = None,
    ignore_author: str = None,
) -> List[Path]:
    """获取目标文件夹下的特定扩展文件列表"""
    file_list = []
    # 兼容带.与不带.
    suffix = suffix.strip().strip(".")
    for flow_path in target.glob(f"**/*.{suffix}"):
        # 递归尝试过滤
        if len(flow_path.parts) - len(target.parts) > depth:
            logger.debug(f"超过递归深度【{depth}】，忽略：{flow_path}")
            continue

        # 仅运行指定流程
        _flow_name = get_flow_name(flow_path)
        if flow_name and _flow_name != flow_name:
            logger.debug(f"不满足指定流程名称【{flow_name}】，忽略：{flow_path}")
            continue

        # 仅运行指定版本的流程
        if flow_version and get_flow_version(flow_path) != flow_version:
            logger.debug(f"版本不满足指定要求【{flow_version}】，忽略：{flow_path}")
            continue

        # 按流程忽略列表过滤——判断是否为忽略流程
        if ignore_path and is_ignore(ignore_path, flow_path):
            logger.debug(f"匹配忽略配置文件，忽略：{flow_path}")
            continue

        # 仅运行指定作者的流程
        if only_author and get_flow_author(flow_path) != only_author:
            logger.debug(f"不满足仅运行指定作者{only_author}的流程，忽略：{flow_path}")
            continue

        # 忽略指定作者的流程
        if ignore_author and get_flow_author(flow_path) == ignore_author:
            logger.debug(f"满足忽略指定作者{only_author}的流程，忽略：{flow_path}")
            continue

        # 忽略包含子流程的流程
        if "子流程" in str(flow_path):
            logger.debug(f"不支持包含子流程的流程，忽略：{flow_path}")
            continue

        # 流程满足以上所有条件，加入待运行列表
        file_list.append(flow_path)

    file_list.sort(reverse=reverse)
    return file_list


def get_flow_version(flow_file_abs_path: Path):
    """获取流程版本"""
    file_utils.check_path_is_exits(flow_file_abs_path, path_type="file")
    flow_version_file_path = flow_file_abs_path.parent / "res" / "version.txt"
    with open(flow_version_file_path, encoding="utf-8") as _file:
        return _file.read().strip()


def get_flow_author(flow_file_abs_path: Path):
    """获取流程作者"""
    file_utils.check_path_is_exits(flow_file_abs_path, path_type="file")
    flow_author_file_path = flow_file_abs_path.parent / "res" / "author.txt"
    with open(flow_author_file_path, encoding="utf-8") as _file:
        return _file.read().strip()


def is_uuid(_uuid: str):
    """判断字符串是否为uuid格式"""
    if len(_uuid) != 36:
        return False

    _e = r"([a-z0-9]{8}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{12})"
    _res = re.match(_e, _uuid)
    if _res:
        return _res.groups()[0]
    else:
        return False


def is_version_no(version: str):
    """判断字符串是否为uuid格式"""
    _e = r"([1-9]{1}.[0-9]{1}.[0-9]{1})"
    _res = re.match(_e, version)
    if _res:
        return _res.groups()[0]
    else:
        return False


def get_flow_name(flow_file_abs_path: Union[str, Path]) -> str:
    """获取流程名称"""
    _flow_name = None
    file_utils.check_path_is_exits(flow_file_abs_path, path_type="file")
    flow_file_abs_path = Path(flow_file_abs_path)
    # 6.0.0版本以前的老流程
    # 所有版本的bot包
    if flow_file_abs_path.stem.lower() != "main":
        _flow_name = flow_file_abs_path.stem
    # 6.0.0及以后版本的流程
    elif flow_file_abs_path.suffix.lower() == ".prj":
        try:
            info_json = file_utils.read_file_content(
                flow_file_abs_path, encoding="utf-8", _return="json"
            )
            _flow_name = info_json["name"]
        except:
            logger.error(f"流程文件解析失败：{flow_file_abs_path}")
            raise
    else:
        # Worker 中运行时bot包缓存目录
        config_path = flow_file_abs_path.parent / "config.json"
        info_json = file_utils.read_file_content(
            config_path, encoding="utf-8", _return="json"
        )
        _flow_name = info_json["name"]
    return _flow_name


def clean_flow(flow_dir_path: Path):
    """清理流程目录中的临时文件"""
    # 清理数据：临时文件，后续生成.bot包时会重新生成这些文件
    file_utils.check_path_is_exits(flow_dir_path, path_type="dir")
    for item in flow_dir_path.iterdir():
        if item.is_file() and item.suffix in [".bot", ".flowc", ".taskc", ".cme", ".bak"]:
            item.unlink()
            continue

        if item.name.lower() in ["log", "tempgit", "config.json", ".git"]:
            if item.is_file():
                item.unlink()
            else:
                file_utils.rm(item)



import os
try:
    from aip import AipOcr
except:
    import time
    os.system("pip install baidu-aip==2.2.18.0")
    time.sleep(2)
    from aip import AipOcr


OCR_APP_ID = os.environ.get("OCR_APP_ID")
OCR_APP_KEY = os.environ.get("OCR_APP_KEY")
OCR_APP_SECRET = os.environ.get("OCR_APP_SECRET")


client = AipOcr(OCR_APP_ID, OCR_APP_KEY, OCR_APP_SECRET)


def ocr(file_bin, accurate: bool = False):
    """OCR识别图片
    :param file_bin: bytes, 文件二进制内容
    :param accurate: bool, 高精度版
    """
    if OCR_APP_ID and OCR_APP_KEY and OCR_APP_SECRET:
        # 通用文字识别（高精度版）
        if accurate:
            data = client.basicAccurate(file_bin)
        # 通用文字识别
        else:
            data = client.basicGeneral(file_bin)
        return data
    else:
        raise ValueError(f"参数没有赋值：OCR_APP_ID，OCR_APP_KEY，OCR_APP_SECRET")



"""工具函数，与打包逻辑无直接关联"""
import math
# coding = utf-8
import os
import uuid
import json
import time
import logging
import inspect
import datetime
import subprocess
from typing import Sequence, Dict
from pathlib import Path
from decimal import Decimal
from functools import wraps


logger = logging.getLogger("DengUtils")


class ExecuteCMDException(Exception):
    """执行外部命令异常"""

    pass


def pop_key_from_dict(my_dict: dict, key, default=None):
    if key in my_dict:
        value = my_dict.pop(key)
    else:
        value = default
    return value


def execute_cmd(
    *popenargs,
    input=None,
    capture_output=True,
    timeout=None,
    check=False,
    level="info",
    encoding="utf-8",
    **kwargs,
):
    """通过subprocess库执行命令行"""
    kwargs["input"] = input
    kwargs["capture_output"] = capture_output
    kwargs["timeout"] = timeout
    kwargs["check"] = check
    ignore_error_log = pop_key_from_dict(kwargs, "ignore_error_log", default=True)
    cmd_output_level = pop_key_from_dict(kwargs, "cmd_output_level", default=level)
    if encoding:
        kwargs["encoding"] = encoding
    if isinstance(popenargs, Sequence):
        if isinstance(popenargs[0], str):
            cmd_text = popenargs[0]
        elif isinstance(popenargs[0], Sequence):
            cmd_text = " ".join(popenargs[0])
        elif isinstance(popenargs[0], Path):
            cmd_text = popenargs[0].name
            kwargs["cwd"] = popenargs[0].parent
        else:
            raise ValueError(f"参数遇到未知情况：{popenargs}")
    else:
        raise ValueError(f"参数遇到未知情况：{popenargs}")

    if level:
        getattr(logger, level)(f"执行命令：{cmd_text}")

    # 增加异常兼容逻辑，处理npm时的可能报错
    for run_count in range(2):
        try:
            _res = subprocess.run(*popenargs, **kwargs)
        except IndexError as e:
            if run_count == 0:
                logger.exception(e)
                logger.warning(
                    f"运行报错，当前capture_output={capture_output}，翻转capture_output参数，再次尝试运行……"
                )
                kwargs["capture_output"] = not kwargs["capture_output"]
            else:
                error_msg = f"执行命令出错：{os.getcwd()} - {cmd_text}\n{str(e)}"
                if not ignore_error_log:
                    logger.error(error_msg)
                raise ExecuteCMDException(error_msg)
        except Exception as e:
            error_msg = f"执行命令出错：{os.getcwd()} - {cmd_text}\n{str(e)}"
            if not ignore_error_log:
                logger.error(error_msg)
            raise ExecuteCMDException(error_msg)
        else:
            if _res.returncode == 0:
                if _res.stdout and cmd_output_level:
                    # 有些命令如npm，会将警告信息打印在stderr中
                    output = f"{byte_to_str(_res.stderr)}\n{byte_to_str(_res.stdout)}"
                    getattr(logger, cmd_output_level)(output)
                return _res
            else:
                error_output = f"{byte_to_str(_res.stderr)}\n{byte_to_str(_res.stdout)}"
                error_msg = f"执行命令出错：{os.getcwd()} - {cmd_text}\n{error_output}"
                if not ignore_error_log:
                    logger.error(error_msg)
                raise ExecuteCMDException(error_msg)


def check_shell_run_result(res_code, desc=""):
    """检查结果，非0时报错"""
    if res_code == 0:
        return True
    else:
        raise ExecuteCMDException(f"{desc}命令执行失败，返回结果={res_code}")


def my_json_serializable(o):
    """补充标准库中json serializable逻辑"""
    if isinstance(o, datetime.datetime):
        return o.strftime("%Y-%m-%d %H:%M:%S")
    if isinstance(o, datetime.date):
        return o.strftime("%Y-%m-%d")
    if isinstance(o, uuid.UUID):
        return str(o)
    if isinstance(o, Path):
        return str(o)
    if isinstance(o, Decimal):
        return float(o)
    if hasattr(o, "__html__"):
        return str(o.__html__())

    try:
        return str(o)
    except Exception:
        raise TypeError(
            f"Object of type {o.__class__.__name__} " f"is not JSON serializable"
        )


class MyJSONEncoder(json.encoder.JSONEncoder):
    """自定义JSON序列化类"""

    def default(self, o):
        return my_json_serializable(o)


def json_to_str(content, indent: int = 4) -> str:
    try:
        return json.dumps(content, ensure_ascii=False, indent=4, cls=MyJSONEncoder)
    except Exception as _error:
        return str(content)


def byte_to_str(src, encoding=None) -> str:
    if isinstance(src, bytes):
        error_list = []
        encoding_list = ["utf-8", "GBK", "GB2312", "GB18030", "ISO-8859-1"]
        if encoding:
            # 已经存在时删除再插入到第一位，确保指定的编码第一个运行
            if encoding in encoding_list:
                encoding_list.remove(encoding)
            encoding_list.insert(encoding, 0)

        for encoding in encoding_list:
            try:
                return str(src, encoding=encoding).strip()
            except UnicodeDecodeError as e:
                error_list.append(e)
        if error_list:
            logger.warning(f"bytes转换成str时出错，尝试的编码有：{encoding_list}，原始对象：{src}")
    elif src is None:
        return ""
    return src.strip() if isinstance(src, str) else str(src)


def bytes_to_str(src, encoding=None):
    return byte_to_str(src, encoding=encoding)


def to_boolean(flag, default=False):
    """将字符串或数字转换成布尔型"""
    if isinstance(flag, bool):
        return flag
    elif flag is None or len(str(flag)) == 0:
        return default
    elif str(flag).lower() in ("false", "no", "0"):
        return False
    else:
        return True


def get_digit_from_input(params=None):
    """从键盘获取一个数字，并做规范性检查
    :param params: Union[List, Tuple],如果给出则输入的数字必须在params内。
    """
    while True:
        num_str = eval(input("请输入一个有效数字："))
        if num_str.isdigit():
            num_int = int(num_str)
            if isinstance(params, (tuple, list)) and len(params) > 0:
                if num_int in params:
                    break
                else:
                    print("输入的不是一个有效选项！")
            else:
                break
        else:
            print("输入的不是一个数字！")
    return num_int


def stat_func_elapsed(func):
    """装饰器：统计被装饰方法的运行耗时"""

    @wraps(func)
    def _stat_func_elapsed(*args, **kwargs):
        func_name = func.__name__
        func_desc = func.__doc__.strip().split()[0] or ""
        logger.info(f"开始运行：{func_name}（{func_desc}）")
        start_time = time.perf_counter()
        try:
            return func(*args, **kwargs)
        finally:
            elapsed = round(time.perf_counter() - start_time, 3)
            logger.info(f"结束运行：{func_name}（{func_desc}），耗时：{elapsed}秒")

    return _stat_func_elapsed


def polishing_int(number: int, length: int, fill_char: str = "0") -> str:
    """数字补齐"""
    number = str(number)
    if len(number) < length:
        return fill_char * (length - len(number)) + number


def extract_func_elapsed(elapsed_collector, parent=None, node=None):
    """装饰器：统计被装饰方法的运行耗时"""

    def _stat_func_elapsed1(func):
        @wraps(func)
        def _stat_func_elapsed2(*args, **kwargs):
            func_name = func.__name__
            func_desc = func.__doc__.strip().split()[0] or ""

            index_no = elapsed_collector.size()
            index_no = polishing_int(index_no, length=3, fill_char="0")

            # 记录打包节点信息
            # 判断节点是否已经存在
            if node in elapsed_collector.nodes.keys():
                node_identifier = node + "-" + str(elapsed_collector.size() + 1)
            elif func_name in elapsed_collector.nodes.keys():
                node_identifier = func_name + "-" + str(elapsed_collector.size() + 1)
            else:
                node_identifier = node or func_name

            elapsed_collector.create_node(
                tag=f"{index_no} - {func_name}（{func_desc}）：运行中",
                identifier=node_identifier,
                parent=parent,
            )

            logger.info(f"开始运行：{func_name}（{func_desc}）")
            start_time = time.perf_counter()
            try:
                return func(*args, **kwargs)
            finally:
                elapsed = round(time.perf_counter() - start_time, 1)
                logger.info(f"结束运行：{func_name}（{func_desc}），耗时：{elapsed}秒")
                elapsed_collector.nodes[
                    node_identifier
                ].tag = f"{index_no} - 耗时【{elapsed}】：{func_name}（{func_desc}）"

        return _stat_func_elapsed2

    return _stat_func_elapsed1


def get_caller_info(depth: int) -> Dict[str, str]:
    """获取调用方名称与描述
    :param depth: int, 函数调用栈递归深度，当前方法为0，依次往上递增
    """
    # 获取调用方描述
    from_obj = inspect.stack()[depth]

    _filename = Path(from_obj.filename).name
    _lineno = from_obj.lineno
    _func_name = from_obj.function

    _func_desc = from_obj.frame.f_code.co_consts[0]
    if _func_desc:
        _func_desc = _func_desc.split("\n")[0].strip()
    else:
        _func_desc = ""
    return {
        "file_abs_path": from_obj.filename,
        "filename": _filename,
        "lineno": _lineno,
        "func_name": _func_name,
        "func_desc": _func_desc,
        "output": f"[{_filename}/{_lineno}/{_func_name}/{_func_desc}]",
    }


def get_caller_desc(depth: int) -> str:
    """获取调用者描述或名称
    :param depth: int, 函数调用栈递归深度，当前方法为0，依次往上递增
    """
    try:
        doc_str = inspect.stack()[depth].frame.f_code.co_consts[0]
        if doc_str:
            doc_str = doc_str.split("\n")[0].strip()
        else:
            doc_str = inspect.stack()[depth].frame.f_code.co_name
        return doc_str
    except Exception as err:
        logger.exception(err)
        return ""


def get_excel_col_name_by_index(col_index: int) -> str:
    """根据列索引获取Excel中的列名"""
    if col_index <= 0:
        raise ValueError("Excel列索引号为从1开始的正整数")

    if 0 < col_index <= 26:
        return chr(col_index + 64)
    else:
        quotient = math.floor(col_index / 26)
        remainder = col_index % 26
        if remainder == 0:
            return get_excel_col_name_by_index(quotient - 1) + "Z"
        else:
            return get_excel_col_name_by_index(quotient) + chr(remainder + 64)
